use pest::Parser;

pub mod build;
pub mod unparse;

// Operators

#[derive(Debug)]
pub enum Operator {
    Binary(BinaryOperator),
    Unary(UnaryOperator),
}

#[derive(Debug)]
pub enum BinaryOperator {
    Or,
    And,
    EqualTo,
    NotEqualTo,
    Greater,
    Less,
    GreaterEqual,
    LessEqual,
    Add,
    Subtract,
    Multiply,
    Divide,
    Modulo,
    LeftShift,
    RightShift,
}

#[derive(Debug)]
pub enum UnaryOperator {
    Not,
    Negative,
    PlusPlus,
    MinusMinus,
    CallOp,
    Spread,
}

// Names

#[derive(Debug, Clone)]
pub struct Identifier {
    pub name: String,
}

#[derive(Debug)]
pub struct Fqn {
    pub identifiers: Vec<Identifier>,
}

// Generic parameters

#[derive(Debug)]
pub struct GenericParameters(pub Vec<GenericParameter>);

impl GenericParameters {
    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }
}

#[derive(Debug)]
pub struct GenericParameter {
    identifier: Identifier,
}

// Generic arguments

#[derive(Debug)]
pub struct GenericArguments(pub Vec<GenericArgument>);

impl GenericArguments {
    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }
}

#[derive(Debug)]
pub struct GenericArgument {
    pub fqn: Fqn,
}

// Type-use and components

#[derive(Debug)]
pub struct TypeUse {
    pub fqn: Fqn,
    pub arguments: Option<TypeArguments>,
}

#[derive(Debug)]
pub enum TypeArguments {
    Generics(GenericArguments),
    Tuple(TupleArguments),
    Function(FunctionTypeArguments),
}

#[derive(Debug)]
pub struct TupleArguments(pub Vec<Fqn>);

impl TupleArguments {
    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }
}

#[derive(Debug)]
pub struct FunctionTypeArguments {
    pub parameters: Parameters,
    pub return_type: Box<TypeUse>,
}

// Function components

#[derive(Debug)]
pub struct Parameters(pub Vec<Parameter>);

impl Parameters {
    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }
}

#[derive(Debug)]
pub struct Parameter {
    identifier: Identifier,
    type_use: TypeUse,
}

#[derive(Debug)]
pub struct ReturnType {
    pub type_use: Box<TypeUse>,
    pub references: References,
}

#[derive(Debug)]
pub struct References(pub Vec<Reference>);

impl References {
    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }
}

#[derive(Debug)]
pub struct Reference {
    pub identifier: Identifier,
}

// Inputs

#[derive(Debug)]
pub enum DelegateOrIdentifier {
    Delegate,
    Identifier(Identifier),
}

#[derive(Debug)]
pub struct InputParameters(pub Vec<DelegateOrIdentifier>);

impl InputParameters {
    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }
}

#[derive(Debug)]
pub struct InputArguments(pub Vec<InputArgument>);

impl InputArguments {
    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }
}

#[derive(Debug)]
pub struct InputArgument {
    pub lhs: DelegateOrIdentifier,
    pub rhs: Identifier
}

// Where guards

#[derive(Debug)]
pub struct WhereGuards(pub Vec<WhereGuard>);

impl WhereGuards {
    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }
}

#[derive(Debug)]
pub struct WhereGuard {
    pub identifier: Identifier,
    pub implements: ImplementsList
}

// Implements

#[derive(Debug)]
pub struct ImplementsList(pub Vec<TypeUse>);

impl ImplementsList {
    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }
}

// Top-level construct

#[derive(Debug)]
pub struct CompilationUnit {
    pub namespace: Option<Fqn>,
    pub declarations: Vec<ModuleLevelDeclaration>,
}

// Declarations allowed in each level

#[derive(Debug)]
pub enum ModuleLevelDeclaration {
    Type(TypeDeclaration),
    Module(ModuleDeclaration),
    Interface(InterfaceDeclaration),
    Class(ClassDeclaration),
    Function(FunctionDeclaration),
    PlatformFunction(PlatformFunctionDeclaration),
}

#[derive(Debug)]
pub enum InterfaceLevelDeclaration {
    Type(TypeDeclaration),
    Module(ModuleDeclaration),
    Interface(InterfaceDeclaration),
    Class(ClassDeclaration),
    Function(InterfaceFunctionDeclaration),
    OperatorFunction(InterfaceOperatorFunctionDeclaration),
}

#[derive(Debug)]
pub enum ClassLevelDeclaration {
    Type(TypeDeclaration),
    Module(ModuleDeclaration),
    Interface(InterfaceDeclaration),
    Class(ClassDeclaration),
    Function(FunctionDeclaration),
    OperatorFunction(OperatorFunctionDeclaration),
    PlatformFunction(PlatformFunctionDeclaration),
    Field(FieldDeclaration),
}

// Declarations

#[derive(Debug)]
pub struct TypeDeclaration {
    pub is_public: bool,
    pub identifier: Identifier,
    pub parameters: Option<TypeParameters>,
    pub where_guards: WhereGuards,
    pub rhs: TypeUse,
}

#[derive(Debug)]
pub struct ModuleDeclaration {
    pub is_public: bool,
    pub identifier: Identifier,
    pub declarations: Vec<ModuleLevelDeclaration>,
}

#[derive(Debug)]
pub struct InterfaceDeclaration {
    pub is_public: bool,
    pub identifier: Identifier,
    pub generics: GenericParameters,
    pub inputs: InputParameters,
    pub implements: ImplementsList,
    pub where_guards: WhereGuards,
    pub declarations: Vec<InterfaceLevelDeclaration>,
}

#[derive(Debug)]
pub struct ClassDeclaration {
    pub is_public: bool,
    pub identifier: Identifier,
    pub generics: GenericParameters,
    pub class_constructor: Option<ClassConstructor>,
    pub inputs: InputArguments,
    pub implements: ImplementsList,
    pub where_guards: WhereGuards,
    pub declarations: Vec<ClassLevelDeclaration>,
}

// Function declarations and components

#[derive(Debug)]
pub enum FunctionModifier {
    Static,
    Cons,
    Mut,
    Ref,
    MutRef,
}

#[derive(Debug)]
pub enum FunctionBody {
    EqualsBody(Expression),
    BlockBody(BlockStatement),
    AliasBody(Identifier),
}

#[derive(Debug)]
pub struct FunctionDeclaration {
    pub is_public: bool,
    pub modifier: Option<FunctionModifier>,
    pub generics: GenericParameters,
    pub identifier: Identifier,
    pub parameters: Parameters,
    pub return_type: Option<ReturnType>,
    pub body: FunctionBody,
}

#[derive(Debug)]
pub struct OperatorFunctionDeclaration {
    pub is_public: bool,
    pub modifier: Option<FunctionModifier>,
    pub generics: GenericParameters,
    pub operator: Operator,
    pub parameters: Parameters,
    pub return_type: Option<ReturnType>,
    pub body: FunctionBody,
}

#[derive(Debug)]
pub struct PlatformFunctionDeclaration {
    pub is_public: bool,
    pub modifier: Option<FunctionModifier>,
    pub generics: GenericParameters,
    pub identifier: Identifier,
    pub parameters: Parameters,
    pub return_type: TypeUse,
}

#[derive(Debug)]
pub struct InterfaceFunctionDeclaration {
    pub modifier: Option<FunctionModifier>,
    pub generics: GenericParameters,
    pub identifier: Identifier,
    pub parameters: Parameters,
    pub return_type: ReturnType,
}

#[derive(Debug)]
pub struct InterfaceOperatorFunctionDeclaration {
    pub modifier: Option<FunctionModifier>,
    pub generics: GenericParameters,
    pub operator: Operator,
    pub parameters: Parameters,
    pub return_type: ReturnType,
}

// Type components

#[derive(Debug)]
pub enum TypeParameters {
    Generic(GenericParameters),
    Tuple(TupleParameters),
    Function(FunctionTypeParameters),
}

#[derive(Debug)]
pub struct TupleParameters(pub Vec<Identifier>);

#[derive(Debug)]
pub struct FunctionTypeParameters {
    pub parameters: Parameters,
    pub return_type: TypeUse,
}

// Class components

#[derive(Debug)]
pub struct ClassConstructor(pub Vec<ClassConstructorParameter>);

#[derive(Debug)]
pub struct ClassConstructorParameter {
    pub is_field: bool,
    pub identifier: Identifier,
    pub declared_type: TypeUse
}

#[derive(Debug)]
pub struct FieldDeclaration {
    pub is_mutable: bool,
    pub identifier: Identifier,
    pub declared_type: TypeUse,
}

// Statements

#[derive(Debug)]
pub struct BlockStatement(pub Vec<Statement>);

#[derive(Debug)]
pub enum Statement {
    BlockStatement(BlockStatement),
    CallStatement(CallExpression),
    VariableDeclarationStatement(VariableDeclarationStatement),
    AssignStatement(AssignStatement),
    ReturnStatement(ReturnStatement),
    IfStatement(IfStatement),
    IfElseStatement(IfElseStatement),
    WhileStatement(WhileStatement),
    ForStatement(ForStatement),
}

#[derive(Debug)]
pub struct VariableDeclarationStatement {
    is_mutable: bool,
    identifier: Identifier,
    declared_type: Option<TypeUse>,
    initializer: Option<Expression>,
}

#[derive(Debug)]
pub struct AssignStatement(pub AssignmentExpression);

#[derive(Debug)]
pub struct ReturnStatement(pub Option<Expression>);

#[derive(Debug)]
pub struct IfStatement {
    pub condition: Expression,
    pub then_branch: BlockStatement,
}

#[derive(Debug)]
pub struct IfElseStatement {
    pub condition: Expression,
    pub then_branch: BlockStatement,
    pub else_ifs: ElseIfs,
    pub else_branch: BlockStatement,
}

#[derive(Debug)]
pub struct ElseIfs(pub Vec<IfStatement>);

#[derive(Debug)]
pub struct WhileStatement {
    pub condition: Expression,
    pub body: BlockStatement,
}

#[derive(Debug)]
pub struct ForStatement {
    pub variable: Identifier,
    pub iterator: Expression,
    pub body: BlockStatement,
}

// Expressions

#[derive(Debug)]
pub enum Expression {
    Binary(BinaryExpression),
    Unary(UnaryExpression),
    Assignment(Box<AssignmentExpression>),
    Call(CallExpression),
    Literal(Literal)
}

#[derive(Debug)]
pub struct BinaryExpression {
    pub left: Box<Expression>,
    pub operator: BinaryOperator,
    pub right: Box<Expression>,
}

#[derive(Debug)]
pub struct UnaryExpression {
    pub operator: UnaryOperator,
    pub expression: Box<Expression>,
}

#[derive(Debug)]
pub struct AssignmentExpression {
    identifier: Identifier,
    expression: Expression,
}

#[derive(Debug)]
pub struct CallExpression {
    pub callee: Box<Expression>,
    pub arguments: CallArguments
}

#[derive(Debug)]
pub struct CallArguments(pub Vec<CallArgument>);

#[derive(Debug)]
pub struct CallArgument(pub Box<Expression>);

#[derive(Debug)]
pub enum Literal {
    IntegerLiteral(i32),
    LongLiteral(i64),
    DoubleLiteral(f64),
    USizeLiteral(usize),
    StringLiteral(String),
    BooleanLiteral(bool),
}
